from __future__ import annotations

from dataclasses import dataclass
from typing import Any, ClassVar

import numpy as np


@dataclass
class builtin_interfaces__msg__Time:
    sec: int
    nanosec: int

    __msgtype__: ClassVar[str] = "builtin_interfaces/msg/Time"


Time = builtin_interfaces__msg__Time


@dataclass
class std_msgs__msg__Header:
    stamp: builtin_interfaces__msg__Time
    frame_id: str

    __msgtype__: ClassVar[str] = "std_msgs/msg/Header"


Header = std_msgs__msg__Header


@dataclass
class sensor_msgs__msg__PointField:
    name: str
    offset: int
    datatype: int
    count: int

    INT8: ClassVar[int] = 1
    UINT8: ClassVar[int] = 2
    INT16: ClassVar[int] = 3
    UINT16: ClassVar[int] = 4
    INT32: ClassVar[int] = 5
    UINT32: ClassVar[int] = 6
    FLOAT32: ClassVar[int] = 7
    FLOAT64: ClassVar[int] = 8

    __msgtype__: ClassVar[str] = "sensor_msgs/msg/PointField"


PointField = sensor_msgs__msg__PointField


@dataclass
class sensor_msgs__msg__PointCloud2:
    header: std_msgs__msg__Header
    height: int
    width: int
    fields: list[sensor_msgs__msg__PointField]
    is_bigendian: bool
    point_step: int
    row_step: int
    data: np.ndarray[Any, np.dtype[np.uint8]]
    is_dense: bool

    __msgtype__: ClassVar[str] = "sensor_msgs/msg/PointCloud2"


PointCloud2 = sensor_msgs__msg__PointCloud2


TYPE_MAPPINGS = [
    (PointField.INT8, np.dtype("int8")),
    (PointField.UINT8, np.dtype("uint8")),
    (PointField.INT16, np.dtype("int16")),
    (PointField.UINT16, np.dtype("uint16")),
    (PointField.INT32, np.dtype("int32")),
    (PointField.UINT32, np.dtype("uint32")),
    (PointField.FLOAT32, np.dtype("float32")),
    (PointField.FLOAT64, np.dtype("float64")),
]

PFTYPE_TO_NPTYPE = dict(TYPE_MAPPINGS)
NPTYPE_TO_PFTYPE = dict((nptype, pftype) for pftype, nptype in TYPE_MAPPINGS)

PFTYPE_SIZES = {
    PointField.INT8: 1,
    PointField.UINT8: 1,
    PointField.INT16: 2,
    PointField.UINT16: 2,
    PointField.INT32: 4,
    PointField.UINT32: 4,
    PointField.FLOAT32: 4,
    PointField.FLOAT64: 8,
}


def build_dtype(msg: PointCloud2) -> list[tuple[str, object]]:
    offset = 0
    dtypes: list[tuple[str, object]] = []
    for field in msg.fields:
        while offset < field.offset:
            dtypes.append((f"__{offset}", np.dtype("uint8")))
            offset += 1

        dtypes.append((field.name, PFTYPE_TO_NPTYPE[field.datatype]))
        offset += PFTYPE_SIZES[field.datatype]

    while offset < msg.point_step:
        dtypes.append((f"__{offset}", np.dtype("uint8")))
        offset += 1

    return dtypes


def pointcloud2_to_array(msg: PointCloud2) -> np.ndarray:
    dtype = build_dtype(msg)

    array = np.frombuffer(msg.data, dtype)
    array = np.hstack([array[n][:, None] for n in array.dtype.names if not n.startswith("__")])

    return array



from pathlib import Path
from tempfile import TemporaryDirectory

import numpy as np
import pytest
from pydantic import ValidationError

from pypcd4 import Encoding, MetaData, PointCloud


def test_parse_pcd_header(pcd_header):
    metadata = MetaData.parse_header(pcd_header)

    assert metadata.version == ".7"
    assert metadata.fields == ("x", "y", "z")
    assert metadata.size == (4, 4, 4)
    assert metadata.type == ("F", "F", "F")
    assert metadata.count == (1, 1, 1)
    assert metadata.width == 213
    assert metadata.height == 1
    assert metadata.viewpoint == (0.0, 0.0, 0.0, 1.0, 0.0, 0.0, 0.0)
    assert metadata.points == 213
    assert metadata.data == "ascii"


def test_parse_pcd_header_random_order(pcd_header_random_order):
    metadata = MetaData.parse_header(pcd_header_random_order)

    assert metadata.version == ".7"
    assert metadata.fields == ("x", "y", "z")
    assert metadata.size == (4, 4, 4)
    assert metadata.type == ("F", "F", "F")
    assert metadata.count == (1, 1, 1)
    assert metadata.width == 213
    assert metadata.height == 1
    assert metadata.viewpoint == (0.0, 0.0, 0.0, 1.0, 0.0, 0.0, 0.0)
    assert metadata.points == 213
    assert metadata.data == "ascii"


def test_parse_pcd_header_with_typical_comments(pcd_header_with_typical_comment):
    metadata = MetaData.parse_header(pcd_header_with_typical_comment)

    assert metadata.version == ".7"
    assert metadata.fields == ("x", "y", "z")
    assert metadata.size == (4, 4, 4)
    assert metadata.type == ("F", "F", "F")
    assert metadata.count == (1, 1, 1)
    assert metadata.width == 213
    assert metadata.height == 1
    assert metadata.viewpoint == (0.0, 0.0, 0.0, 1.0, 0.0, 0.0, 0.0)
    assert metadata.points == 213
    assert metadata.data == "ascii"


def test_parse_pcd_header_with_random_comments(pcd_header_with_random_comments):
    metadata = MetaData.parse_header(pcd_header_with_random_comments)

    assert metadata.version == ".7"
    assert metadata.fields == ("x", "y", "z")
    assert metadata.size == (4, 4, 4)
    assert metadata.type == ("F", "F", "F")
    assert metadata.count == (1, 1, 1)
    assert metadata.width == 213
    assert metadata.height == 1
    assert metadata.viewpoint == (0.0, 0.0, 0.0, 1.0, 0.0, 0.0, 0.0)
    assert metadata.points == 213
    assert metadata.data == "ascii"


def test_parse_pcd_header_with_empty_lines(pcd_header_with_empty_lines):
    metadata = MetaData.parse_header(pcd_header_with_empty_lines)

    assert metadata.version == ".7"
    assert metadata.fields == ("x", "y", "z")
    assert metadata.size == (4, 4, 4)
    assert metadata.type == ("F", "F", "F")
    assert metadata.count == (1, 1, 1)
    assert metadata.width == 213
    assert metadata.height == 1
    assert metadata.viewpoint == (0.0, 0.0, 0.0, 1.0, 0.0, 0.0, 0.0)
    assert metadata.points == 213
    assert metadata.data == "ascii"


def test_parse_pcd_header_with_no_points(pcd_header_with_no_points):
    metadata = MetaData.parse_header(pcd_header_with_no_points)

    assert metadata.version == ".7"
    assert metadata.fields == ("x", "y", "z")
    assert metadata.size == (4, 4, 4)
    assert metadata.type == ("F", "F", "F")
    assert metadata.count == (1, 1, 1)
    assert metadata.width == 0
    assert metadata.height == 0
    assert metadata.viewpoint == (0.0, 0.0, 0.0, 1.0, 0.0, 0.0, 0.0)
    assert metadata.points == 0
    assert metadata.data == "ascii"


def test_parse_pcd_header_with_multi_count_fields(pcd_header_with_multi_count_fields):
    metadata = MetaData.parse_header(pcd_header_with_multi_count_fields)

    assert metadata.version == ".7"
    assert metadata.fields == ("x", "y", "z")
    assert metadata.size == (4, 4, 4)
    assert metadata.type == ("F", "F", "F")
    assert metadata.count == (1, 1, 4)
    assert metadata.width == 213
    assert metadata.height == 1
    assert metadata.viewpoint == (0.0, 0.0, 0.0, 1.0, 0.0, 0.0, 0.0)
    assert metadata.points == 213
    assert metadata.data == "ascii"


def test_parse_pcd_header_with_negative_viewpoint(pcd_header_with_negative_viewpoint):
    metadata = MetaData.parse_header(pcd_header_with_negative_viewpoint)

    assert metadata.version == ".7"
    assert metadata.fields == ("x", "y", "z")
    assert metadata.size == (4, 4, 4)
    assert metadata.type == ("F", "F", "F")
    assert metadata.count == (1, 1, 1)
    assert metadata.width == 213
    assert metadata.height == 1
    assert metadata.viewpoint == (1.0, -2.0, 3.0, 1.0, -1.0, 2.0, -3.0)
    assert metadata.points == 213
    assert metadata.data == "ascii"


def test_parse_pcd_header_with_underscore_in_fields(pcd_header_with_underscore_in_fields):
    metadata = MetaData.parse_header(pcd_header_with_underscore_in_fields)

    assert metadata.version == ".7"
    assert metadata.fields != ("_", "y", "_")
    assert metadata.size == (4, 4, 4)
    assert metadata.type == ("F", "F", "F")
    assert metadata.count == (1, 1, 1)
    assert metadata.width == 213
    assert metadata.height == 1
    assert metadata.viewpoint == (0.0, 0.0, 0.0, 1.0, 0.0, 0.0, 0.0)
    assert metadata.points == 213
    assert metadata.data == "ascii"


def test_parse_pcd_header_with_extra_items(pcd_header_with_extra_items):
    metadata = MetaData.parse_header(pcd_header_with_extra_items)

    assert metadata.version == ".7"
    assert metadata.fields == ("x", "y", "z")
    assert metadata.size == (4, 4, 4)
    assert metadata.type == ("F", "F", "F")
    assert metadata.count == (1, 1, 1)
    assert metadata.width == 213
    assert metadata.height == 1
    assert metadata.viewpoint == (0.0, 0.0, 0.0, 1.0, 0.0, 0.0, 0.0)
    assert metadata.points == 213
    assert metadata.data == "ascii"


def test_compose_pcd_header(pcd_header):
    metadata = MetaData.parse_header(pcd_header)

    assert metadata.compose_header() == (
        "VERSION .7\nFIELDS x y z\nSIZE 4 4 4\nTYPE F F F\nCOUNT 1 1 1\n"
        "WIDTH 213\nHEIGHT 1\nVIEWPOINT 0.0 0.0 0.0 1.0 0.0 0.0 0.0\nPOINTS 213\nDATA ascii\n"
    )


def test_build_dtype(pcd_header):
    metadata = MetaData.parse_header(pcd_header)

    assert metadata.build_dtype() == [("x", "<f4"), ("y", "<f4"), ("z", "<f4")]


def test_build_dtype_with_multi_count_fields(pcd_header_with_multi_count_fields):
    metadata = MetaData.parse_header(pcd_header_with_multi_count_fields)

    assert metadata.build_dtype() == [
        ("x", "<f4"),
        ("y", "<f4"),
        ("z_0000", "<f4"),
        ("z_0001", "<f4"),
        ("z_0002", "<f4"),
        ("z_0003", "<f4"),
    ]


def test_parse_pcd_header_with_invalid_version(pcd_header_with_invalid_version):
    with pytest.raises(ValidationError):
        MetaData.parse_header(pcd_header_with_invalid_version)


def test_parse_pcd_header_with_invalid_field_type(pcd_header_with_invalid_field_type):
    with pytest.raises(ValidationError):
        MetaData.parse_header(pcd_header_with_invalid_field_type)


def test_parse_pcd_header_with_invalid_data_type(pcd_header_with_invalid_data_type):
    with pytest.raises(ValidationError):
        MetaData.parse_header(pcd_header_with_invalid_data_type)


def test_parse_pcd_header_with_invalid_viewpoint(pcd_header_with_invalid_viewpoint):
    with pytest.raises(ValidationError):
        MetaData.parse_header(pcd_header_with_invalid_viewpoint)


def test_parse_pcd_header_with_negative_points(pcd_header_with_negative_points):
    with pytest.raises(ValidationError):
        MetaData.parse_header(pcd_header_with_negative_points)


def test_parse_pcd_header_with_missing_items(pcd_header_with_missing_items):
    with pytest.raises(ValidationError):
        MetaData.parse_header(pcd_header_with_missing_items)


def test_load_ascii_pcd(xyzrgb_ascii_path):
    pc = PointCloud.from_path(xyzrgb_ascii_path)

    assert pc.metadata.data == "ascii"
    assert pc.pc_data.dtype.names == pc.metadata.fields
    assert len(pc.pc_data) == pc.metadata.points


def test_load_xyzrgb_ascii_with_underscore_pcd(xyzrgb_ascii_with_underscore_path):
    pc = PointCloud.from_path(xyzrgb_ascii_with_underscore_path)

    assert pc.metadata.data == "ascii"
    assert pc.pc_data.dtype.names == pc.metadata.fields
    assert len(pc.pc_data) == pc.metadata.points


def test_load_xyzrgb_ascii_with_empty_points_pcd(xyzrgb_ascii_with_empty_points_path):
    pc = PointCloud.from_path(xyzrgb_ascii_with_empty_points_path)

    assert pc.metadata.data == "ascii"
    assert pc.pc_data.dtype.names == pc.metadata.fields
    assert len(pc.pc_data) == pc.metadata.points


def test_load_binary_pcd(xyzrgb_binary_path):
    pc = PointCloud.from_path(xyzrgb_binary_path)

    assert pc.metadata.data == "binary"
    assert pc.pc_data.dtype.names == pc.metadata.fields
    assert len(pc.pc_data) == pc.metadata.points


def test_load_binary_compressed_pcd(xyzrgb_binary_compressed_path):
    pc = PointCloud.from_path(xyzrgb_binary_compressed_path)

    assert pc.metadata.data == "binary_compressed"
    assert pc.pc_data.dtype.names == pc.metadata.fields
    assert len(pc.pc_data) == pc.metadata.points


def test_from_points():
    array = np.array([[1, 2, 3], [4, 5, 6]])
    fields = ("x", "y", "z")
    types = (np.float32, np.float32, np.float32)
    count = (1, 1, 1)

    pc = PointCloud.from_points(array, fields, types, count)

    assert pc.fields == fields
    assert pc.types == types
    assert pc.points == 2
    assert pc.pc_data.dtype.names == fields

    assert np.array_equal(pc.pc_data["x"], array.T[0])
    assert np.array_equal(pc.pc_data["y"], array.T[1])
    assert np.array_equal(pc.pc_data["z"], array.T[2])

    assert pc.pc_data["x"].dtype == np.dtype(types[0])
    assert pc.pc_data["y"].dtype == np.dtype(types[1])
    assert pc.pc_data["z"].dtype == np.dtype(types[2])


def test_from_points_with_invalid_points():
    array = "invalid_points"
    fields = ("x", "y", "z")
    types = (np.float32, np.float32, np.float32)

    with pytest.raises(TypeError):
        PointCloud.from_points(array, fields, types)


def test_from_points_with_invalid_fields_and_types():
    points = np.random.random((100, 4))
    fields = ("x", "y", "z", "a")
    types = (np.float32, np.float32)

    with pytest.raises(ValueError):
        PointCloud.from_points(points, fields, types)


def test_from_xyz_points():
    points = np.random.random((100, 3))

    pc = PointCloud.from_xyz_points(points)

    assert pc.fields == ("x", "y", "z")
    assert pc.types == (np.float32, np.float32, np.float32)


def test_from_xyzi_points():
    points = np.random.random((100, 4))

    pc = PointCloud.from_xyzi_points(points)

    assert pc.fields == ("x", "y", "z", "intensity")
    assert pc.types == (np.float32, np.float32, np.float32, np.float32)


def test_from_xyzl_points():
    points = np.random.random((100, 4))

    label_type = np.int32
    pc = PointCloud.from_xyzl_points(points, label_type)

    assert pc.fields == ("x", "y", "z", "label")
    assert pc.types == (np.float32, np.float32, np.float32, label_type)

    label_type = np.float32
    pc = PointCloud.from_xyzl_points(points, label_type)

    assert pc.fields == ("x", "y", "z", "label")
    assert pc.types == (np.float32, np.float32, np.float32, label_type)

    label_type = np.uint8
    pc = PointCloud.from_xyzl_points(points, label_type)

    assert pc.fields == ("x", "y", "z", "label")
    assert pc.types == (np.float32, np.float32, np.float32, label_type)


def test_from_xyzrgb_points():
    points = np.random.random((100, 4))

    pc = PointCloud.from_xyzrgb_points(points)

    assert pc.fields == ("x", "y", "z", "rgb")
    assert pc.types == (np.float32, np.float32, np.float32, np.float32)


def test_from_xyzrgbl_points():
    points = np.random.random((100, 5))

    label_type = np.int32
    pc = PointCloud.from_xyzrgbl_points(points, label_type)

    assert pc.fields == ("x", "y", "z", "rgb", "label")
    assert pc.types == (np.float32, np.float32, np.float32, np.float32, label_type)

    label_type = np.float32
    pc = PointCloud.from_xyzrgbl_points(points, label_type)

    assert pc.fields == ("x", "y", "z", "rgb", "label")
    assert pc.types == (np.float32, np.float32, np.float32, np.float32, label_type)

    label_type = np.uint8
    pc = PointCloud.from_xyzrgbl_points(points, label_type)

    assert pc.fields == ("x", "y", "z", "rgb", "label")
    assert pc.types == (np.float32, np.float32, np.float32, np.float32, label_type)


def test_from_xyzil_points():
    points = np.random.random((100, 5))

    label_type = np.int32
    pc = PointCloud.from_xyzil_points(points, label_type)

    assert pc.fields == ("x", "y", "z", "intensity", "label")
    assert pc.types == (np.float32, np.float32, np.float32, np.float32, label_type)

    label_type = np.float32
    pc = PointCloud.from_xyzil_points(points, label_type)

    assert pc.fields == ("x", "y", "z", "intensity", "label")
    assert pc.types == (np.float32, np.float32, np.float32, np.float32, label_type)

    label_type = np.uint8
    pc = PointCloud.from_xyzil_points(points, label_type)

    assert pc.fields == ("x", "y", "z", "intensity", "label")
    assert pc.types == (np.float32, np.float32, np.float32, np.float32, label_type)


def test_from_xyzirgb_points():
    points = np.random.random((100, 5))

    pc = PointCloud.from_xyzirgb_points(points)

    assert pc.fields == ("x", "y", "z", "intensity", "rgb")
    assert pc.types == (np.float32, np.float32, np.float32, np.float32, np.float32)


def test_from_xyzirgbl_points():
    points = np.random.random((100, 6))

    label_type = np.int32
    pc = PointCloud.from_xyzirgbl_points(points, label_type)

    assert pc.fields == ("x", "y", "z", "intensity", "rgb", "label")
    assert pc.types == (np.float32, np.float32, np.float32, np.float32, np.float32, label_type)

    label_type = np.float32
    pc = PointCloud.from_xyzirgbl_points(points, label_type)

    assert pc.fields == ("x", "y", "z", "intensity", "rgb", "label")
    assert pc.types == (np.float32, np.float32, np.float32, np.float32, np.float32, label_type)

    label_type = np.uint8
    pc = PointCloud.from_xyzirgbl_points(points, label_type)

    assert pc.fields == ("x", "y", "z", "intensity", "rgb", "label")
    assert pc.types == (np.float32, np.float32, np.float32, np.float32, np.float32, label_type)


def test_from_xyzt_points():
    points = np.random.random((100, 5))

    pc = PointCloud.from_xyzt_points(points)

    assert pc.fields == ("x", "y", "z", "sec", "nsec")
    assert pc.types == (np.float32, np.float32, np.float32, np.uint32, np.uint32)


def test_from_xyzir_points():
    points = np.random.random((100, 5))

    pc = PointCloud.from_xyzir_points(points)

    assert pc.fields == ("x", "y", "z", "intensity", "ring")
    assert pc.types == (np.float32, np.float32, np.float32, np.float32, np.uint16)


def test_from_xyzirt_points():
    points = np.random.random((100, 6))

    pc = PointCloud.from_xyzirt_points(points)

    assert pc.fields == ("x", "y", "z", "intensity", "ring", "time")
    assert pc.types == (np.float32, np.float32, np.float32, np.float32, np.uint16, np.float32)


def test_from_xyzit_points():
    points = np.random.random((100, 5))

    pc = PointCloud.from_xyzit_points(points)

    assert pc.fields == ("x", "y", "z", "intensity", "timestamp")
    assert pc.types == (np.float32, np.float32, np.float32, np.float32, np.float64)


def test_from_xyzis_points():
    points = np.random.random((100, 5))

    pc = PointCloud.from_xyzis_points(points)

    assert pc.fields == ("x", "y", "z", "intensity", "stamp")
    assert pc.types == (np.float32, np.float32, np.float32, np.float32, np.float64)


def test_from_xyzisc_points():
    points = np.random.random((100, 6))

    pc = PointCloud.from_xyzisc_points(points)

    assert pc.fields == ("x", "y", "z", "intensity", "stamp", "classification")
    assert pc.types == (np.float32, np.float32, np.float32, np.float32, np.float64, np.uint8)


def test_from_xyzrgbs_points():
    points = np.random.random((100, 5))

    pc = PointCloud.from_xyzrgbs_points(points)

    assert pc.fields == ("x", "y", "z", "rgb", "stamp")
    assert pc.types == (np.float32, np.float32, np.float32, np.float32, np.float64)


def test_from_xyzirgbs_points():
    points = np.random.random((100, 6))

    pc = PointCloud.from_xyzirgbs_points(points)

    assert pc.fields == ("x", "y", "z", "intensity", "rgb", "stamp")
    assert pc.types == (np.float32, np.float32, np.float32, np.float32, np.float32, np.float64)


def test_from_xyzirgbsc_points():
    points = np.random.random((100, 7))

    pc = PointCloud.from_xyzirgbsc_points(points)

    assert pc.fields == ("x", "y", "z", "intensity", "rgb", "stamp", "classification")
    assert pc.types == (
        np.float32,
        np.float32,
        np.float32,
        np.float32,
        np.float32,
        np.float64,
        np.uint8,
    )


def test_from_xyziradt_points():
    points = np.random.random((100, 9))

    pc = PointCloud.from_xyziradt_points(points)

    assert pc.fields == (
        "x",
        "y",
        "z",
        "intensity",
        "ring",
        "azimuth",
        "distance",
        "return_type",
        "time_stamp",
    )
    assert pc.types == (
        np.float32,
        np.float32,
        np.float32,
        np.float32,
        np.uint16,
        np.float32,
        np.float32,
        np.uint8,
        np.float64,
    )


def test_from_ouster_points():
    points = np.random.random((100, 9))

    pc = PointCloud.from_ouster_points(points)

    assert pc.fields == (
        "x",
        "y",
        "z",
        "intensity",
        "t",
        "reflectivity",
        "ring",
        "ambient",
        "range",
    )
    assert pc.types == (
        np.float32,
        np.float32,
        np.float32,
        np.float32,
        np.float32,
        np.uint16,
        np.uint8,
        np.uint16,
        np.uint32,
    )


def test_from_invalid_points():
    points = np.random.random((100, 1))

    with pytest.raises(ValueError):
        PointCloud.from_ouster_points(points)


def test_encode_rgb():
    r = np.array([249, 249, 249])
    g = np.array([230, 230, 230])
    b = np.array([206, 206, 206])

    expect = np.array((r << 16) | (g << 8) | (b << 0), np.uint32)
    expect.dtype = np.float32

    output = PointCloud.encode_rgb((r, g, b))

    assert np.allclose(output, expect)


def test_decode_rgb():
    r = 249
    g = 230
    b = 206

    expect = np.array([[r, g, b], [r, g, b]])
    output = PointCloud.decode_rgb(
        np.array([[(r << 16) | (g << 8) | (b << 0)], [(r << 16) | (g << 8) | (b << 0)]])
    )

    assert np.allclose(output, expect)


def test_numpy():
    in_points = np.random.random((100, 3))
    fields = ("x", "y", "z")
    types = (np.float32, np.float32, np.float32)
    pc = PointCloud.from_points(in_points, fields, types)

    out_points = pc.numpy()
    assert np.allclose(out_points, in_points)

    out_points = pc.numpy(fields=("x", "y"))
    assert np.allclose(out_points, in_points[:, :2])

    out_points = pc.numpy(fields=[])
    assert np.allclose(out_points, np.empty((0, 0)))


def test_save_as_ascii():
    in_points = np.random.randint(0, 1000, (100, 3))
    fields = ("x", "y", "z")
    types = (np.float32, np.int8, np.uint64)
    pc = PointCloud.from_points(in_points, fields, types)

    with TemporaryDirectory() as dirname:
        path = dirname / Path("test.pcd")

        pc.save(path, Encoding.ASCII)
        assert path.is_file()

        pc2 = PointCloud.from_path(path)
        assert np.allclose(pc2.numpy(), pc.numpy())
        assert pc2.fields == pc.fields
        assert pc2.types == pc.types
        assert pc2.points == pc.points
        assert pc2.metadata.data == pc.metadata.data


def test_save_as_binary():
    in_points = np.random.randint(0, 1000, (100, 3))
    fields = ("x", "y", "z")
    types = (np.float32, np.int8, np.uint64)
    pc = PointCloud.from_points(in_points, fields, types)

    with TemporaryDirectory() as dirname:
        path = dirname / Path("test.pcd")

        pc.save(path, Encoding.BINARY)
        assert path.is_file()

        pc2 = PointCloud.from_path(path)
        assert np.allclose(pc2.numpy(), pc.numpy())
        assert pc2.fields == pc.fields
        assert pc2.types == pc.types
        assert pc2.points == pc.points
        assert pc2.metadata.data == pc.metadata.data


def test_save_as_binary_compressed():
    in_points = np.random.randint(0, 1000, (100, 3))
    fields = ("x", "y", "z")
    types = (np.float32, np.int8, np.uint64)
    pc = PointCloud.from_points(in_points, fields, types)

    with TemporaryDirectory() as dirname:
        path = dirname / Path("test.pcd")

        pc.save(path, Encoding.BINARY_COMPRESSED)
        assert path.is_file()

        pc2 = PointCloud.from_path(path)
        assert np.allclose(pc2.numpy(), pc.numpy())
        assert pc2.fields == pc.fields
        assert pc2.types == pc.types
        assert pc2.points == pc.points
        assert pc2.metadata.data == pc.metadata.data


def test_save_with_empty_points():
    in_points = np.empty((0, 3))
    fields = ("x", "y", "z")
    types = (np.float32, np.float32, np.float32)
    pc = PointCloud.from_points(in_points, fields, types)

    with TemporaryDirectory() as dirname:
        path = dirname / Path("test.pcd")

        pc.save(path)
        assert path.is_file()

        pc2 = PointCloud.from_path(path)
        assert np.allclose(pc2.numpy(), pc.numpy())
        assert pc2.fields == pc.fields
        assert pc2.types == pc.types
        assert pc2.points == pc.points



from pathlib import Path

import pytest


@pytest.fixture
def pcd_header():
    return [
        "VERSION .7",
        "FIELDS x y z",
        "SIZE 4 4 4",
        "TYPE F F F",
        "COUNT 1 1 1",
        "WIDTH 213",
        "HEIGHT 1",
        "VIEWPOINT 0 0 0 1 0 0 0",
        "POINTS 213",
        "DATA ascii",
    ]


@pytest.fixture
def pcd_header_random_order():
    return [
        "VIEWPOINT 0 0 0 1 0 0 0",
        "TYPE F F F",
        "COUNT 1 1 1",
        "FIELDS x y z",
        "WIDTH 213",
        "VERSION .7",
        "HEIGHT 1",
        "DATA ascii",
        "POINTS 213",
        "SIZE 4 4 4",
    ]


@pytest.fixture
def pcd_header_with_typical_comment():
    return [
        "# .PCD v.7 - Point Cloud Data file format",
        "VIEWPOINT 0 0 0 1 0 0 0",
        "TYPE F F F",
        "COUNT 1 1 1",
        "FIELDS x y z",
        "WIDTH 213",
        "VERSION .7",
        "HEIGHT 1",
        "DATA ascii",
        "POINTS 213",
        "SIZE 4 4 4",
    ]


@pytest.fixture
def pcd_header_with_random_comments():
    return [
        "# comment",
        "VIEWPOINT 0 0 0 1 0 0 0",
        "### comment",
        "TYPE F F F",
        "COUNT 1 1 1",
        "# comment",
        "# comment",
        "# comment",
        "FIELDS x y z",
        "WIDTH 213",
        "# comment ###",
        "VERSION .7",
        "HEIGHT 1",
        "DATA ascii",
        "# comment ###",
        "POINTS 213",
        "SIZE 4 4 4",
        "# comment",
    ]


@pytest.fixture
def pcd_header_with_empty_lines():
    return [
        "",
        "VERSION .7",
        "FIELDS x y z",
        "",
        "",
        "SIZE 4 4 4",
        "TYPE F F F",
        "COUNT 1 1 1",
        "",
        "WIDTH 213",
        "HEIGHT 1",
        "VIEWPOINT 0 0 0 1 0 0 0",
        "POINTS 213",
        "DATA ascii",
        "",
        "",
        "",
    ]


@pytest.fixture
def pcd_header_with_no_points():
    return [
        "VERSION .7",
        "FIELDS x y z",
        "SIZE 4 4 4",
        "TYPE F F F",
        "COUNT 1 1 1",
        "WIDTH 0",
        "HEIGHT 0",
        "VIEWPOINT 0 0 0 1 0 0 0",
        "POINTS 0",
        "DATA ascii",
    ]


@pytest.fixture
def pcd_header_with_multi_count_fields():
    return [
        "VERSION .7",
        "FIELDS x y z",
        "SIZE 4 4 4",
        "TYPE F F F",
        "COUNT 1 1 4",
        "WIDTH 213",
        "HEIGHT 1",
        "VIEWPOINT 0 0 0 1 0 0 0",
        "POINTS 213",
        "DATA ascii",
    ]


@pytest.fixture
def pcd_header_with_negative_viewpoint():
    return [
        "VERSION .7",
        "FIELDS x y z",
        "SIZE 4 4 4",
        "TYPE F F F",
        "COUNT 1 1 1",
        "WIDTH 213",
        "HEIGHT 1",
        "VIEWPOINT 1 -2 3 1 -1 2 -3",
        "POINTS 213",
        "DATA ascii",
    ]


@pytest.fixture
def pcd_header_with_underscore_in_fields():
    return [
        "VERSION .7",
        "FIELDS _ y _",
        "SIZE 4 4 4",
        "TYPE F F F",
        "COUNT 1 1 1",
        "WIDTH 213",
        "HEIGHT 1",
        "VIEWPOINT 0 0 0 1 0 0 0",
        "POINTS 213",
        "DATA ascii",
    ]


@pytest.fixture
def pcd_header_with_extra_items():
    return [
        "VERSION .7",
        "FIELDS x y z",
        "SIZE 4 4 4",
        "TYPE F F F",
        "COUNT 1 1 1",
        "WIDTH 213",
        "HEIGHT 1",
        "VIEWPOINT 0 0 0 1 0 0 0",
        "POINTS 213",
        "DATA ascii",
        "COLUMN a b c",
        "EXTRA hope do nothing",
    ]


@pytest.fixture
def pcd_header_with_invalid_version():
    return [
        "VERSION 999",
        "FIELDS x y z",
        "SIZE 4 4 4",
        "TYPE F F F",
        "COUNT 1 1 1",
        "WIDTH 213",
        "HEIGHT 1",
        "VIEWPOINT 0 0 0 1 0 0 0",
        "POINTS 213",
        "DATA ascii",
    ]


@pytest.fixture
def pcd_header_with_invalid_field_type():
    return [
        "VERSION .7",
        "FIELDS x y z",
        "SIZE 4 4 4",
        "TYPE U S A",
        "COUNT 1 1 1",
        "WIDTH 213",
        "HEIGHT 1",
        "VIEWPOINT 0 0 0 1 0 0 0",
        "POINTS 213",
        "DATA ascii",
    ]


@pytest.fixture
def pcd_header_with_invalid_data_type():
    return [
        "VERSION .7",
        "FIELDS x y z",
        "SIZE 4 4 4",
        "TYPE F F F",
        "COUNT 1 1 1",
        "WIDTH 213",
        "HEIGHT 1",
        "VIEWPOINT 0 0 0 1 0 0 0",
        "POINTS 213",
        "DATA iicsa",
    ]


@pytest.fixture
def pcd_header_with_invalid_viewpoint():
    return [
        "VERSION .7",
        "FIELDS x y z",
        "SIZE 4 4 4",
        "TYPE F F F",
        "COUNT 1 1 1",
        "WIDTH 213",
        "HEIGHT 1",
        "VIEWPOINT 0",
        "POINTS 213",
        "DATA ascii",
    ]


@pytest.fixture
def pcd_header_with_negative_points():
    return [
        "VERSION .7",
        "FIELDS x y z",
        "SIZE 4 4 4",
        "TYPE F F F",
        "COUNT 1 1 1",
        "WIDTH 0",
        "HEIGHT 0",
        "VIEWPOINT 0 0 0 1 0 0 0",
        "POINTS -123",
        "DATA ascii",
    ]


@pytest.fixture
def pcd_header_with_missing_items():
    return [
        "VERSION .7",
        "FIELDS x y z",
        # "SIZE 4 4 4",
        "TYPE F F F",
        "COUNT 1 1 1",
        "WIDTH 213",
        # "HEIGHT 1",
        "VIEWPOINT 0 0 0 1 0 0 0",
        # "POINTS 213",
        "DATA ascii",
    ]


@pytest.fixture
def xyzrgb_ascii_path():
    return f"{Path(__file__).resolve().parent}/pcd/ascii.pcd"


@pytest.fixture
def xyzrgb_ascii_with_underscore_path():
    return f"{Path(__file__).resolve().parent}/pcd/ascii_with_underscore.pcd"


@pytest.fixture
def xyzrgb_ascii_with_empty_points_path():
    return f"{Path(__file__).resolve().parent}/pcd/ascii_with_empty_points.pcd"


@pytest.fixture
def xyzrgb_binary_path():
    return f"{Path(__file__).resolve().parent}/pcd/binary.pcd"


@pytest.fixture
def xyzrgb_binary_compressed_path():
    return f"{Path(__file__).resolve().parent}/pcd/binary_compressed.pcd"



from __future__ import annotations

import random
import re
import string
import struct
from enum import Enum
from pathlib import Path
from typing import BinaryIO, List, Literal, Optional, Sequence, Tuple, Type, Union

import lzf
import numpy as np
from pydantic import BaseModel, NonNegativeInt, PositiveInt
from pydantic.dataclasses import dataclass

from .pointcloud2 import PFTYPE_TO_NPTYPE, PointCloud2, pointcloud2_to_array

NpNumberType = Type[Union[np.integer, np.floating]]
PathLike = Union[str, Path]

MetaDataVersion = Literal[".7", "0.7"]
MetaDataViewPoint = Tuple[float, float, float, float, float, float, float]

NUMPY_TYPE_TO_PCD_TYPE: dict[
    np.dtype,
    Tuple[str, int],
] = {
    np.dtype("uint8"): ("U", 1),
    np.dtype("uint16"): ("U", 2),
    np.dtype("uint32"): ("U", 4),
    np.dtype("uint64"): ("U", 8),
    np.dtype("int8"): ("I", 1),
    np.dtype("int16"): ("I", 2),
    np.dtype("int32"): ("I", 4),
    np.dtype("int64"): ("I", 8),
    np.dtype("float32"): ("F", 4),
    np.dtype("float64"): ("F", 8),
}

PCD_TYPE_TO_NUMPY_TYPE: dict[
    Tuple[str, int],
    NpNumberType,
] = {
    ("F", 4): np.float32,
    ("F", 8): np.float64,
    ("U", 1): np.uint8,
    ("U", 2): np.uint16,
    ("U", 4): np.uint32,
    ("U", 8): np.uint64,
    ("I", 1): np.int8,
    ("I", 2): np.int16,
    ("I", 4): np.int32,
    ("I", 8): np.int64,
}

HEADER_PATTERN = re.compile(r"(\w+)\s+([\w\s\.\-?\d+\.?\d*]+)")


class Encoding(str, Enum):
    ASCII = "ascii"
    BINARY = "binary"
    BINARY_COMPRESSED = "binary_compressed"
    BINARYSCOMPRESSED = "binaryscompressed"


@dataclass
class MetaData(BaseModel):
    fields: Tuple[str, ...]
    size: Tuple[PositiveInt, ...]
    type: Tuple[Literal["F", "U", "I"], ...]
    count: Tuple[PositiveInt, ...]
    points: NonNegativeInt
    width: NonNegativeInt
    height: NonNegativeInt = 1
    version: MetaDataVersion = "0.7"
    viewpoint: MetaDataViewPoint = (0.0, 0.0, 0.0, 1.0, 0.0, 0.0, 0.0)
    data: Encoding = Encoding.BINARY_COMPRESSED

    @staticmethod
    def parse_header(lines: List[str]) -> MetaData:
        """
        Parse a header in accordance with the PCD (Point Cloud Data) file format.
        See https://pointclouds.org/documentation/tutorials/pcd_file_format.html.

        Args:
            lines (List[str]): A list of lines of the header

        Raises:
            pydantic.ValidationError: If the header is invalid

        Returns:
            MetaData: Parsed MetaData object
        """

        _header = {}
        for line in lines:
            if line.startswith("#") or len(line) < 2:
                continue

            if (match := re.match(HEADER_PATTERN, line)) is None:
                continue

            value = match.group(2).split()
            if (key := match.group(1).lower()) in ["version", "data"]:
                _header[key] = value[0]
            elif key in ["width", "height", "points"]:
                _header[key] = int(value[0])
            elif key in ["fields"]:
                _header[key] = tuple(
                    [
                        "#$%&~~"
                        + "".join(random.choices(string.ascii_letters + string.digits, k=6))
                        if s == "_"
                        else s
                        for s in value
                    ]
                )
            elif key in ["type"]:
                _header[key] = tuple(value)
            elif key in ["size", "count"]:
                _header[key] = tuple(int(v) for v in value)
            elif key in ["viewpoint"]:
                _header[key] = tuple(float(v) for v in value)
            else:
                pass

        return MetaData.model_validate(_header)

    def compose_header(self) -> str:
        header: List[str] = []

        header.append(f"VERSION {self.version}")
        header.append(
            f"FIELDS {' '.join(['_' if s.startswith('#$%&~~') else s for s in self.fields])}"
        )
        header.append(f"SIZE {' '.join([str(v) for v in self.size])}")
        header.append(f"TYPE {' '.join(self.type)}")
        header.append(f"COUNT {' '.join([str(v) for v in self.count])}")
        header.append(f"WIDTH {self.width}")
        header.append(f"HEIGHT {self.height}")
        header.append(f"VIEWPOINT {' '.join([str(v) for v in self.viewpoint])}")
        header.append(f"POINTS {self.points}")
        header.append(f"DATA {self.data.value}")

        return "\n".join(header) + "\n"

    def build_dtype(self) -> np.dtype[np.void]:
        field_names: List[str] = []
        type_names: List[np.dtype] = []

        for i, field in enumerate(self.fields):
            np_type: np.dtype = np.dtype(PCD_TYPE_TO_NUMPY_TYPE[(self.type[i], self.size[i])])

            if (count := self.count[i]) == 1:
                field_names.append(field)
                type_names.append(np_type)
            else:
                field_names.extend([f"{field}_{i:04d}" for i in range(count)])
                type_names.extend([np_type] * count)

        return np.dtype([x for x in zip(field_names, type_names)])


def _parse_pc_data(fp: BinaryIO, metadata: MetaData) -> np.ndarray:
    dtype = metadata.build_dtype()

    if metadata.points > 0:
        if metadata.data == Encoding.ASCII:
            pc_data = np.loadtxt(fp, dtype, delimiter=" ")
        elif metadata.data == Encoding.BINARY:
            buffer = fp.read(metadata.points * dtype.itemsize)
            pc_data = np.frombuffer(buffer, dtype)  # type: ignore
        else:
            compressed_size, uncompressed_size = struct.unpack("II", fp.read(8))  # type: ignore

            buffer = lzf.decompress(fp.read(compressed_size), uncompressed_size)
            if (actual_size := len(buffer)) != uncompressed_size:
                raise RuntimeError(
                    f"Failed to decompress data. "
                    f"Expected decompressed file size is {uncompressed_size}, but got {actual_size}"
                )

            offset = 0
            pc_data = np.zeros(metadata.width, dtype=dtype)
            for name in dtype.names:
                dt: np.dtype = dtype[name]
                bytes = dt.itemsize * metadata.width
                pc_data[name] = np.frombuffer(buffer[offset : (offset + bytes)], dtype=dt)
                offset += bytes
    else:
        pc_data = np.empty((0, len(metadata.fields)), dtype)

    return pc_data


def _compose_pc_data(
    points: Union[np.ndarray, Sequence[np.ndarray]], metadata: MetaData
) -> np.ndarray:
    arrays: Sequence[np.ndarray] = tuple(points.T) if isinstance(points, np.ndarray) else points

    return np.rec.fromarrays(arrays, dtype=metadata.build_dtype())


class PointCloud:
    def __init__(self, metadata: MetaData, pc_data: np.ndarray) -> None:
        self.metadata = metadata
        self.pc_data = pc_data

    @staticmethod
    def from_fileobj(fp: BinaryIO) -> PointCloud:
        lines: List[str] = []
        while True:
            line = fp.readline().strip()
            lines.append(line.decode(encoding="utf-8") if isinstance(line, bytes) else line)

            if lines[-1].startswith("DATA"):
                break

        metadata = MetaData.parse_header(lines)
        pc_data = _parse_pc_data(fp, metadata)

        return PointCloud(metadata, pc_data)

    @staticmethod
    def from_path(path: PathLike) -> PointCloud:
        """Create PointCloud from file path

        Args:
            path (PathLike): File path

        Raises:
            FileNotFoundError: Raise if `path` does not exist.

        Returns:
            PointCloud: PointCloud object

        >>> PointCloud.from_path("test.pcd")
        <pypcd4.PointCloud object at 0x7f7c9a3d8b00>
        """

        with open(path, mode="rb") as fp:
            return PointCloud.from_fileobj(fp)

    @staticmethod
    def from_points(
        points: Union[np.ndarray, List[np.ndarray], Tuple[np.ndarray]],
        fields: Sequence[str],
        types: Sequence[Union[NpNumberType, np.dtype]],
        count: Optional[Sequence[int]] = None,
    ) -> PointCloud:
        """Create PointCloud from 2D numpy array or sequence of 1D numpy arrays of each field

        Args:
            points (Union[np.ndarray, Sequence[np.ndarray]]): Numpy array
            fields (Sequence[str]): Field names for each numpy array or sequence of numpy arrays
            types (Sequence[Union[NpNumberType, np.dtype]]): Numpy type for each numpy array or
                sequence of numpy arrays
            count (Optional[Sequence[int]], optional): Number of values for each numpy array or
                sequence of numpy arrays. If None, set to 1 for all fields. Defaults to None.

        Raises:
            TypeError: Raise if `points` type is neither `np.ndarray` nor `Sequence[np.ndarray]`.

        Returns:
            PointCloud: PointCloud object

        >>> PointCloud.from_points(
                np.array([[1, 2, 3], [4, 5, 6]]),
                ("x", "y", "z"),
                (np.float32, np.float32, np.float32)
            )
        <pypcd4.PointCloud object at 0x7f7c9a3d8b00>

        >>> PointCloud.from_points(
                [
                    np.array([1, 2, 3]),
                    np.array([4, 5, 6])
                ],
                ("x", "y", "z"),
                (np.float32, np.float32, np.float32)
            )
        <pypcd4.PointCloud object at 0x7f7c9a3d8b00>
        """

        if not isinstance(points, (np.ndarray, list, tuple)):
            raise TypeError(
                "`points` must be a `np.ndarray`, `List[np.ndarray]` or `Tuple[np.ndarray]` type, "
                f"but got `{type(points).__name__}` type"
            )

        if len(fields) != len(types):
            raise ValueError(
                f"`fields` and `types` must have the same length, "
                f"but got {len(fields)} fields and {len(types)} types"
            )

        num_fields = points.shape[1] if isinstance(points, np.ndarray) else len(points)
        if num_fields != len(fields):
            raise ValueError(
                f"`points` must have {len(fields)} fields {fields}, " f"but got {num_fields} fields"
            )

        if count is None:
            count = [1] * len(tuple(fields))

        pcd_types, sizes = [], []
        for type_ in types:
            t, s = NUMPY_TYPE_TO_PCD_TYPE[np.dtype(type_)]
            pcd_types.append(t)
            sizes.append(s)

        num_points = len(points) if isinstance(points, np.ndarray) else len(points[0])

        metadata = MetaData.model_validate(
            {
                "fields": fields,
                "size": sizes,
                "type": pcd_types,
                "count": count,
                "width": num_points,
                "points": num_points,
            }
        )

        return PointCloud(metadata, _compose_pc_data(points, metadata))

    @staticmethod
    def from_xyz_points(points: np.ndarray) -> PointCloud:
        fields = ("x", "y", "z")
        types = (np.float32, np.float32, np.float32)

        return PointCloud.from_points(points, fields, types)

    @staticmethod
    def from_xyzi_points(points: np.ndarray) -> PointCloud:
        fields = ("x", "y", "z", "intensity")
        types = (np.float32, np.float32, np.float32, np.float32)

        return PointCloud.from_points(points, fields, types)

    @staticmethod
    def from_xyzl_points(points: np.ndarray, label_type: NpNumberType = np.float32) -> PointCloud:
        fields = ("x", "y", "z", "label")
        types = (np.float32, np.float32, np.float32, label_type)

        return PointCloud.from_points(points, fields, types)

    @staticmethod
    def from_xyzrgb_points(points: np.ndarray) -> PointCloud:
        fields = ("x", "y", "z", "rgb")
        types = (np.float32, np.float32, np.float32, np.float32)

        return PointCloud.from_points(points, fields, types)

    @staticmethod
    def from_xyzrgbl_points(
        points: np.ndarray, label_type: NpNumberType = np.float32
    ) -> PointCloud:
        fields = ("x", "y", "z", "rgb", "label")
        types = (np.float32, np.float32, np.float32, np.float32, label_type)

        return PointCloud.from_points(points, fields, types)

    @staticmethod
    def from_xyzil_points(points: np.ndarray, label_type: NpNumberType = np.float32) -> PointCloud:
        fields = ("x", "y", "z", "intensity", "label")
        types = (np.float32, np.float32, np.float32, np.float32, label_type)

        return PointCloud.from_points(points, fields, types)

    @staticmethod
    def from_xyzirgb_points(points: np.ndarray) -> PointCloud:
        fields = ("x", "y", "z", "intensity", "rgb")
        types = (np.float32, np.float32, np.float32, np.float32, np.float32)

        return PointCloud.from_points(points, fields, types)

    @staticmethod
    def from_xyzirgbl_points(
        points: np.ndarray, label_type: NpNumberType = np.float32
    ) -> PointCloud:
        fields = ("x", "y", "z", "intensity", "rgb", "label")
        types = (np.float32, np.float32, np.float32, np.float32, np.float32, label_type)

        return PointCloud.from_points(points, fields, types)

    @staticmethod
    def from_xyzt_points(points: np.ndarray) -> PointCloud:
        fields = ("x", "y", "z", "sec", "nsec")
        types = (np.float32, np.float32, np.float32, np.uint32, np.uint32)

        return PointCloud.from_points(points, fields, types)

    @staticmethod
    def from_xyzir_points(points: np.ndarray) -> PointCloud:
        fields = ("x", "y", "z", "intensity", "ring")
        types = (np.float32, np.float32, np.float32, np.float32, np.uint16)

        return PointCloud.from_points(points, fields, types)

    @staticmethod
    def from_xyzirt_points(points: np.ndarray) -> PointCloud:
        fields = ("x", "y", "z", "intensity", "ring", "time")
        types = (np.float32, np.float32, np.float32, np.float32, np.uint16, np.float32)

        return PointCloud.from_points(points, fields, types)

    @staticmethod
    def from_xyzit_points(points: np.ndarray) -> PointCloud:
        fields = ("x", "y", "z", "intensity", "timestamp")
        types = (np.float32, np.float32, np.float32, np.float32, np.float64)

        return PointCloud.from_points(points, fields, types)

    @staticmethod
    def from_xyzis_points(points: np.ndarray) -> PointCloud:
        fields = ("x", "y", "z", "intensity", "stamp")
        types = (np.float32, np.float32, np.float32, np.float32, np.float64)

        return PointCloud.from_points(points, fields, types)

    @staticmethod
    def from_xyzisc_points(points: np.ndarray) -> PointCloud:
        fields = ("x", "y", "z", "intensity", "stamp", "classification")
        types = (np.float32, np.float32, np.float32, np.float32, np.float64, np.uint8)

        return PointCloud.from_points(points, fields, types)

    @staticmethod
    def from_xyzrgbs_points(points: np.ndarray) -> PointCloud:
        fields = ("x", "y", "z", "rgb", "stamp")
        types = (np.float32, np.float32, np.float32, np.float32, np.float64)

        return PointCloud.from_points(points, fields, types)

    @staticmethod
    def from_xyzirgbs_points(points: np.ndarray) -> PointCloud:
        fields = ("x", "y", "z", "intensity", "rgb", "stamp")
        types = (
            np.float32,
            np.float32,
            np.float32,
            np.float32,
            np.float32,
            np.float64,
        )

        return PointCloud.from_points(points, fields, types)

    @staticmethod
    def from_xyzirgbsc_points(points: np.ndarray) -> PointCloud:
        fields = ("x", "y", "z", "intensity", "rgb", "stamp", "classification")
        types = (
            np.float32,
            np.float32,
            np.float32,
            np.float32,
            np.float32,
            np.float64,
            np.uint8,
        )

        return PointCloud.from_points(points, fields, types)

    @staticmethod
    def from_xyziradt_points(points: np.ndarray) -> PointCloud:
        fields = (
            "x",
            "y",
            "z",
            "intensity",
            "ring",
            "azimuth",
            "distance",
            "return_type",
            "time_stamp",
        )
        types = (
            np.float32,
            np.float32,
            np.float32,
            np.float32,
            np.uint16,
            np.float32,
            np.float32,
            np.uint8,
            np.float64,
        )

        return PointCloud.from_points(points, fields, types)

    @staticmethod
    def from_ouster_points(points: np.ndarray) -> PointCloud:
        fields = (
            "x",
            "y",
            "z",
            "intensity",
            "t",
            "reflectivity",
            "ring",
            "ambient",
            "range",
        )
        types = (
            np.float32,
            np.float32,
            np.float32,
            np.float32,
            np.float32,
            np.uint16,
            np.uint8,
            np.uint16,
            np.uint32,
        )

        return PointCloud.from_points(points, fields, types)

    @staticmethod
    def from_msg(msg: PointCloud2) -> PointCloud:
        points = pointcloud2_to_array(msg)
        fields = [f.name for f in msg.fields]
        types = [PFTYPE_TO_NPTYPE[f.datatype] for f in msg.fields]

        return PointCloud.from_points(points, fields, types)  # type: ignore

    @staticmethod
    def encode_rgb(rgb: np.ndarray | list[np.ndarray]) -> np.ndarray:
        """
        Encode Nx3 uint8 array with RGB values to
        Nx1 float32 array with bit-packed RGB
        """

        if isinstance(rgb, np.ndarray):
            if rgb.ndim == 1:
                rgb = rgb[:, None]
        else:
            rgb = np.hstack([v[:, None] if v.ndim == 1 else v for v in rgb])

        rgb_u32 = rgb.astype(np.uint32)
        rgb_u32 = np.array(
            (rgb_u32[:, 0] << 16) | (rgb_u32[:, 1] << 8) | (rgb_u32[:, 2] << 0), dtype=np.uint32
        )
        rgb_u32.dtype = np.float32  # type: ignore

        return rgb_u32

    @staticmethod
    def decode_rgb(rgb: np.ndarray) -> np.ndarray:
        """
        Decode Nx1 float32 array with bit-packed RGB to
        Nx3 uint8 array with RGB values
        """

        rgb_u32 = rgb.astype(np.uint32)
        r = np.asarray((rgb_u32 >> 16) & 255, dtype=np.uint8)
        g = np.asarray((rgb_u32 >> 8) & 255, dtype=np.uint8)
        b = np.asarray(rgb_u32 & 255, dtype=np.uint8)

        return np.hstack((r, g, b))

    @property
    def fields(self) -> Tuple[str, ...]:
        """
        Returns fields of the point cloud

        Returns:
            Tuple[str, ...]: Tuple of field names for each field

        >>> pc.fields
        ("x", "y", "z")
        """

        return self.metadata.fields

    @property
    def types(self) -> Tuple[NpNumberType, ...]:
        """Returns types of the point cloud

        Returns:
            Tuple[NpNumberType, ...]: Tuple of numpy types for each field

        >>> pc.types
        (np.float32, np.float32, np.float32)
        """

        return tuple(
            PCD_TYPE_TO_NUMPY_TYPE[ts] for ts in zip(self.metadata.type, self.metadata.size)
        )

    @property
    def points(self) -> int:
        """Returns number of points in the point cloud

        Returns:
            int: Tuple of number of points for each field

        >>> pc.points
        1000
        """

        return self.metadata.points

    def numpy(self, fields: Optional[Sequence[str]] = None) -> np.ndarray:
        """Returns numpy array of the point cloud

        Args:
            fields (Optional[Sequence[str]], optional): Fields to return.
                If None, all fields are returned. Defaults to None.

        Returns:
            np.ndarray: Numpy array of the point cloud

        >>> pc.fields
        ("x", "y", "z")

        >>> pc.numpy()
        array([[0.0, 0.1, 0.2],
               [0.3, 0.4, 0.5],
               ...,
               [0.9, 1.0, 1.1]])

        >>> pc.numpy(("x", "y"))
        array([[0.0, 0.1],
               [0.3, 0.4],
               ...,
               [0.9, 1.0]])

        >>> pc.numpy(("x", "y", "z"))
        array([[0.0, 0.1, 0.2],
               [0.3, 0.4, 0.5],
               ...,
               [0.9, 1.0, 1.1]])
        """

        if fields is None:
            fields = self.fields

        if len(fields) == 0:
            return np.empty((0, 0))

        if self.metadata.points == 0:
            return np.empty((0, len(fields)))

        _stack = tuple(self.pc_data[field] for field in fields)

        return np.vstack(_stack).T

    def _save_as_ascii(self, fp: BinaryIO) -> None:
        """Saves point cloud to a file as a ascii

        Args:
            path (PathLike): Path to the file
        """

        format = []
        for type, count in zip(self.metadata.type, self.metadata.count):
            if type == "F":
                format.extend(["%.10f"] * count)
            elif type == "I":
                format.extend(["%d"] * count)
            else:
                format.extend(["%u"] * count)

        np.savetxt(fp, self.pc_data, format)

    def _save_as_binary(self, fp: BinaryIO) -> None:
        """Saves point cloud to a file as a binary

        Args:
            path (PathLike): Path to the file
        """

        fp.write(self.pc_data.tobytes())

    def _save_as_binary_compressed(self, fp: BinaryIO) -> None:
        """Saves point cloud to a file as a binary compressed

        Args:
            path (PathLike): Path to the file
        """

        uncompressed = b"".join(
            np.ascontiguousarray(self.pc_data[field]).tobytes()
            for field in self.pc_data.dtype.names
        )

        if (compressed := lzf.compress(uncompressed)) is None:
            compressed = uncompressed

        fp.write(struct.pack("II", len(compressed), len(uncompressed)))
        fp.write(compressed)

    def save(self, path: PathLike, encoding: Encoding = Encoding.BINARY) -> None:
        """Saves point cloud to a file

        Args:
            path (PathLike): Path to the file
            encoding (Encoding, optional): Encoding to use. Defaults to Encoding.BINARY_COMPRESSED.
        """

        self.metadata.data = encoding

        with open(path, mode="wb") as fp:
            header = self.metadata.compose_header().encode()
            fp.write(header)

            if self.metadata.points < 1:
                return

            if encoding == Encoding.ASCII:
                self._save_as_ascii(fp)
            elif encoding == Encoding.BINARY:
                self._save_as_binary(fp)
            else:
                self._save_as_binary_compressed(fp)
